/**
 * Copyright (c) 2004, Christoph Leisegang
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.

 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Created on 09.07.2004
 *
 */

package com.linkwerk.util;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * <b>Regexp</b> Regular Expressions covers extension functions that provide 
 * facilities to do with regular expressions to XSLT.
 * <p>
 * This package implements the regular expression extension following the
 * rules given by {@link <a href="http://exslt.org/regexp/index.html">exslt.org</a>}.
 * <p>
 * The regular expressions used in this module use the 
 * regular expression syntax of the <code>java.util.regex package</code> of 
 * the <b>Java 2 Platform, Standard Edition (since 1.4)</b>, 
 * and are <b>not</b> following the Javascript regular expression syntax. 
 * 
 * 
 * @author Christoph Leisegang
 * @version 1.0
 * 
 * @see <code>java.util.regex</code> package of 
 * the <b>Java 2 Platform, Standard Edition (since 1.4)</b>
 */
public abstract class Regexp {
    
	/**
	 * Concatenates the text nodes of element-nodes and 
	 * values of text-, attribute-, comment and 
	 * processing-instruction-nodes of the given node list
	 * 
	 * @param nl the node list to be concatenated
	 * @return the concatenated string
	 */
	private static String nlToString(NodeList nl){
		StringBuffer buf = new StringBuffer(); 

		for (int i=0; i < nl.getLength(); i++){
		    switch(nl.item(i).getNodeType()) {
		    	case Node.ELEMENT_NODE :
		    	    buf.append(nlToString(nl.item(i).getChildNodes()));
		    	    break;
		    	case Node.TEXT_NODE :
		    	case Node.ATTRIBUTE_NODE :
		    	case Node.COMMENT_NODE :
		    	case Node.PROCESSING_INSTRUCTION_NODE :
					buf.append(nl.item(i).getNodeValue());
					break;
		    	default : break;
		    }
		} 
		return buf.toString(); 
	}
	
	/**
	 * The <code>test</code> function returns true if the node list given 
	 * as the first parameter matches the regular expression
	 * given as the second parameter. 
	 * <p>
	 * The input sequence is scanned looking for the first subsequence 
	 * that matches the pattern.
	 * <p>
	 * All the text nodes or attribute values of the node list 
	 * are concatenated before the test is executed.
	 * 
	 * @see #test(String, String, String) test(java.lang.String input, java.lang.String pattern, java.lang.String flags)
	 * 
	 * @param nl the node list to be tested
	 * @param pattern the regular expression
	 * @return <b>true</b>, if the node list matches the regular expression
	 */
	public static boolean test(NodeList nl, String pattern)
	{
		return test(nlToString(nl), pattern, "");
	}

	/**
	 * The <code>test</code> function returns true if the node list given 
	 * as the first parameter matches the regular expression
	 * given as the second parameter. 	 
	 * <p>
	 * The input sequence is scanned looking for the first subsequence 
	 * that matches the pattern.
	 * <p>
	 * All the text nodes or attribute values of the node list 
	 * are concatenated before the test is executed.
	 * 
	 * @see #test(String, String, String) test(java.lang.String input, java.lang.String pattern, java.lang.String flags)
	 * 
	 * @param nl the node list to be tested
	 * @param pattern the regular expression
	 * @param flags the flags to be used ("g","i")
	 * @return <b>true</b>, if the node list matches the regular expression
	 */
	public static boolean test(NodeList nl, String pattern, String flags)
	{
		return test(nlToString(nl), pattern, flags);
	}

	/**
	 * The <code>test</code> function returns true if the string given 
	 * as the first parameter matches the regular expression
	 * given as the second parameter.
 	 * <p>
	 * The input sequence is scanned looking for the first subsequence 
	 * that matches the pattern.
	 * <p>
	 * The second parameter is a regular expression that 
	 * follows the regular expression syntax of the 
	 * <code>java.util.regex</code> package of the  
	 * <b>Java 2 Platform, Standard Edition (since 1.4)</b>.
	 * 
	 * @see #test(String, String, String) test(java.lang.String input, java.lang.String pattern, java.lang.String flags)
	 * 
	 * @param input the string to be tested
	 * @param pattern the regular expression
	 * @return <b>true</b>, if the node list matches the regular expression
	 */
	public static boolean test(String input, String pattern)
	{
		return test(input, pattern, "");
	}
	
	/**
	 * The <code>test</code> function returns true if the string given 
	 * as the first parameter matches the regular expression
	 * given as the second parameter.
	 * <p>
	 * The input sequence is scanned looking for the first subsequence 
	 * that matches the pattern.
	 * <p>
	 * The second parameter is a regular expression that 
	 * follows the regular expression syntax of the 
	 * <code>java.util.regex</code> package of the  
	 * <b>Java 2 Platform, Standard Edition (since 1.4)</b>.
	 * <p>
	 * The third parameter is a string consisting of flags
	 * to be used by <code>test</code>. If a character is present than
	 * the flag is true. Possible flags are:
	 * <p>
	 * <b>g:</b> global test - has no effect on this function and
	 * is just retained for consistency with <code>Regexp.match</code>
	 * and <code>Regexp.replace</code>.
	 * <b>i:</b> case insenitive - the regular expression is treated
	 * as case insensitiv. The default uses case senitive regular 
	 * expressions. 
	 * 
	 * @param input the string to be tested
	 * @param pattern the regular expression
	 * @param flags the flags to be used ("g","i")
	 * @return <b>true</b>, if the node list matches the regular expression
	 */
	public static boolean test(String input, String pattern, String flags)
	{
 		int patternFlags = (flags.indexOf("i") >= 0 ? Pattern.CASE_INSENSITIVE : 0);

 		String inputStr = (input == null ? "" : input);
 		String patternStr = (pattern == null ? "" : pattern);
 		
		Pattern p = Pattern.compile(patternStr, patternFlags);
		Matcher matcher = p.matcher(inputStr);
		return matcher.find();
	}

	/**
	 * The <code>match</code> function returns a node set of
	 * match elements, for each string value that is equal to a
	 * portion of the first parameter node list, captured by the
	 * second parameter regular expression
	 * 
	 * All the text nodes or attributes of the elements or 
	 * attribute values of the node list are concatenated
	 * before the match is executed.
	 * 
	 * @see #match(String, String, String) match(java.lang.String input, java.lang.String pattern, java.lang.String flags)
	 * 
	 * @param nl the node list to be matched
	 * @param pattern the regular expression
	 * @return node set of match elements
	 */
	public static NodeList match(NodeList nl, String pattern)
	{
		return match(nlToString(nl), pattern, "");
	}

	/**
	 * The <code>match</code> function returns a node set of
	 * match elements, for each string value that is equal to a
	 * portion of the first parameter node list, captured by the
	 * second parameter regular expression.
	 * 
	 * All the text nodes or attribute values of the node list 
	 * are concatenated before match is executed.
	 * 
	 * @see #match(String, String, String) match(java.lang.String input, java.lang.String pattern, java.lang.String flags)
	 * 
	 * @param nl the node list to be matched
	 * @param pattern the regular expression
	 * @param flags the flags to be used ("g","i")
	 * @return node set of match elements
	 */
	public static NodeList match(NodeList nl, String pattern, String flags)
	{
		return match(nlToString(nl), pattern, flags);
	}

	/**
	 * The <code>match</code> function returns a node set of
	 * match elements, for each string value that is equal to a
	 * portion of the first parameter string, captured by the
	 * second parameter regular expression.
	 * 
	 * @see #match(String, String, String) match(java.lang.String input, java.lang.String pattern, java.lang.String flags)
	 * 
	 * @param input the string to be matched
	 * @param pattern the regular expression
	 * @return node set of match elements
	 */
	public static NodeList match(String input, String pattern)
	{
		return match(input, pattern, "");
	}

	/**
	 * The <code>match</code> function returns a node set of
	 * match elements, for each string value that is equal to a
	 * portion of the first parameter string, captured by the
	 * second parameter regular expression
	 * <p>
	 * The second parameter is a regular expression that 
	 * follows the regular expression syntax of the 
	 * <code>java.util.regex</code> package of the  
	 * <b>Java 2 Platform, Standard Edition (since 1.4)</b>.
	 * <p>
	 * The third parameter is a string consisting of flags
	 * to be used by <code>match</code>. If a character is present than
	 * the flag is true. Possible flags are:
	 * <p>
	 * <b>g:</b> global match - the submatches of all the matches
	 * in the string are returned. The default (character is not present)
	 * returns only the submatches of the first match in the string.
	 * <b>i:</b> case insenitive - the regular expression is treated
	 * as case insensitiv. The default uses case senitive regular 
	 * expressions.
	 * 
	 * @param input the string to be matched
	 * @param pattern the regular expression
	 * @param flags the flags to be used ("g","i")
	 * @return node set of match elements
	 */
	public static NodeList match(String input, String pattern, String flags)
	{
		int patternFlags = (flags.indexOf("i") >= 0 ? Pattern.CASE_INSENSITIVE : 0);

 		String inputStr = (input == null ? "" : input);
 		String patternStr = (pattern == null ? "" : pattern);
 		
 		Pattern p = Pattern.compile(patternStr, patternFlags);
		Matcher matcher = p.matcher(inputStr);

		Document doc = null;
		StringBuffer buf = new StringBuffer();

		buf.append("<groups>");
	
		if(flags.indexOf("g") >= 0) {
		    while(matcher.find()) {
				for (int i=1; i <= matcher.groupCount(); i++){
					buf.append("<group>");
					buf.append(matcher.group(i) == null ? "" : matcher.group(i));
					buf.append("</group>");
				}
		    }
		} else {
		    if(matcher.find()) {
				for (int i=0; i <= matcher.groupCount(); i++){
					buf.append("<group>");
					buf.append(matcher.group(i) == null ? "" : matcher.group(i));
					buf.append("</group>");
				}
		    }
		}
		
		buf.append("</groups>");

	    try
		{
		    doc = DOM.parse(new ByteArrayInputStream(buf.toString().getBytes()));
		}
		catch(IOException e) {}
		catch(SAXException e) {}
        
		return doc.getElementsByTagName("group");
	}
	
	/**
	 * The <code>replace</code> function returns a string with parts
	 * of the string given as the first parameter that match
	 * second parameter regular expression replaced with another string.
	 * 
	 * @see #replace(String, String, String, String) replace(java.lang.String input, java.lang.String pattern, java.lang.String flags java.lang.String repStr)
	 * 
	 * @param nl the node list to be matched and replaced
	 * @param pattern the regular expression
	 * @param flags the flags to be used ("g","i")
	 * @param replace the replacing string
	 * @return the replaced string 
	 */
	public static String replace(NodeList nl, String pattern, String flags, String replace)
	{
		return replace(nlToString(nl), pattern, flags, replace);
	}

	/**
	 * The <code>replace</code> function returns a string with parts
	 * of the string given as the first parameter that match
	 * second parameter regular expression replaced with another string.
	 * <p>
	 * The second parameter is a regular expression that 
	 * follows the regular expression syntax of the 
	 * <code>java.util.regex</code> package of the  
	 * <b>Java 2 Platform, Standard Edition (since 1.4)</b>.
	 * <p>
	 * The third parameter is a string consisting of flags
	 * to be used by <code>replace</replace>. If a character is present than
	 * the flag is true. Possible flags are:
	 * <p>
	 * <b>g:</b> global match - the submatches of all the matches
	 * in the string are returned. The default (character is not present)
	 * returns only the submatches of the first match in the string.
	 * <b>i:</b> case insenitive - the regular expression is treated
	 * as case insensitiv. The default uses case senitive regular 
	 * expressions.
	 * <p>
	 * The fourth parameter is the string to replace the matched parts. 
	 * 
	 * @param input the string to be matched and replaced
	 * @param pattern the regular expression
	 * @param flags the flags to be used ("g","i")
	 * @param replace the replacing string
	 * @return the replaced string 
	 */
	public static String replace(String input, String pattern, String flags, String replace)
	{
		int patternFlags = (flags.indexOf("i") >= 0 ? Pattern.CASE_INSENSITIVE : 0);

		String inputStr = (input == null ? "" : input);
 		String patternStr = (pattern == null ? "" : pattern);
 		String replaceStr = (replace == null ? "" : replace);

		Pattern p = Pattern.compile(patternStr, patternFlags);
		Matcher matcher = p.matcher(inputStr);
		return (flags.indexOf("g") >= 0 ? matcher.replaceAll(replaceStr) : matcher.replaceFirst(replaceStr));
	}

}
